# Generated by Django 3.1.7 on 2021-10-12 10:39

from django.conf import settings
from django.contrib.postgres.operations import CITextExtension, TrigramExtension
import django.contrib.postgres.fields
import django.contrib.postgres.fields.citext
import django.contrib.postgres.search
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import fernet_fields.fields
import resolwe.flow.models.fields
import uuid
import versionfield.fields


class Migration(migrations.Migration):

    replaces = [
        ("flow", "0001_squashed_0030_change_slug_field"),
        ("flow", "0002_set_process_owners"),
        ("flow", "0003_data_dependency_1"),
        ("flow", "0004_data_dependency_2"),
        ("flow", "0005_data_dependency_3"),
        ("flow", "0006_add_total_size"),
        ("flow", "0007_data_size"),
        ("flow", "0008_compute_size"),
        ("flow", "0009_make_size_mandatory"),
        ("flow", "0010_add_secret"),
        ("flow", "0011_preserve_parents"),
        ("flow", "0012_recreate_empty_parents"),
        ("flow", "0013_migration_history"),
        ("flow", "0014_track_resources"),
        ("flow", "0015_refactor_relations_1"),
        ("flow", "0016_refactor_relations_2"),
        ("flow", "0017_refactor_relations_3"),
        ("flow", "0018_process_is_active"),
        ("flow", "0019_relation_type_cicharfield"),
        ("flow", "0020_process_category_default"),
        ("flow", "0021_collection_tags"),
        ("flow", "0022_process_entity_1"),
        ("flow", "0023_process_entity_2"),
        ("flow", "0024_process_entity_3"),
        ("flow", "0025_entity_type"),
        ("flow", "0026_data_scheduled"),
        ("flow", "0027_data_purged"),
        ("flow", "0028_add_data_location"),
        ("flow", "0029_storage_m2m"),
        ("flow", "0030_add_data_duplicate"),
        ("flow", "0031_add_entity_duplicate"),
        ("flow", "0032_add_collection_duplicate"),
        ("flow", "0033_move_purged"),
        ("flow", "0034_callable_defaults"),
        ("flow", "0035_process_entity_new"),
        ("flow", "0036_add_m2o_fields"),
        ("flow", "0037_migrate_m2o"),
        ("flow", "0038_remove_m2m_fields"),
        ("flow", "0039_entity_collection_cascade"),
        ("flow", "0040_remove_entity_descriptor_completed"),
        ("flow", "0041_remove_download_perm"),
        ("flow", "0042_delete_obsolete_perms"),
        ("flow", "0043_full_text_search"),
    ]

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        CITextExtension(),
        TrigramExtension(),
        migrations.CreateModel(
            name="Collection",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.fields.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                ("description", models.TextField(blank=True)),
                ("settings", models.JSONField(default=dict)),
                ("descriptor", models.JSONField(default=dict)),
                ("descriptor_dirty", models.BooleanField(default=False)),
                (
                    "tags",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        default=list,
                        size=None,
                    ),
                ),
                ("search", django.contrib.postgres.search.SearchVectorField(null=True)),
                ("duplicated", models.DateTimeField(blank=True, null=True)),
            ],
            options={
                "permissions": (
                    ("view_collection", "Can view collection"),
                    ("edit_collection", "Can edit collection"),
                    ("share_collection", "Can share collection"),
                    ("owner_collection", "Is owner of the collection"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Data",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.fields.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "scheduled",
                    models.DateTimeField(blank=True, db_index=True, null=True),
                ),
                ("started", models.DateTimeField(blank=True, db_index=True, null=True)),
                (
                    "finished",
                    models.DateTimeField(blank=True, db_index=True, null=True),
                ),
                ("duplicated", models.DateTimeField(blank=True, null=True)),
                (
                    "checksum",
                    models.CharField(
                        db_index=True,
                        max_length=64,
                        validators=[
                            django.core.validators.RegexValidator(
                                code="invalid_checksum",
                                message="Checksum is exactly 40 alphanumerics",
                                regex="^[0-9a-f]{64}$",
                            )
                        ],
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("UP", "Uploading"),
                            ("RE", "Resolving"),
                            ("WT", "Waiting"),
                            ("PP", "Preparing"),
                            ("PR", "Processing"),
                            ("OK", "Done"),
                            ("ER", "Error"),
                            ("DR", "Dirty"),
                        ],
                        default="RE",
                        max_length=2,
                    ),
                ),
                ("process_pid", models.PositiveIntegerField(blank=True, null=True)),
                ("process_progress", models.PositiveSmallIntegerField(default=0)),
                ("process_rc", models.PositiveSmallIntegerField(blank=True, null=True)),
                (
                    "process_info",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        default=list,
                        size=None,
                    ),
                ),
                (
                    "process_warning",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        default=list,
                        size=None,
                    ),
                ),
                (
                    "process_error",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        default=list,
                        size=None,
                    ),
                ),
                ("input", models.JSONField(default=dict)),
                ("output", models.JSONField(default=dict)),
                ("size", models.BigIntegerField()),
                ("descriptor", models.JSONField(default=dict)),
                ("descriptor_dirty", models.BooleanField(default=False)),
                ("named_by_user", models.BooleanField(default=False)),
                (
                    "tags",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        default=list,
                        size=None,
                    ),
                ),
                ("process_memory", models.PositiveIntegerField(default=0)),
                ("process_cores", models.PositiveSmallIntegerField(default=0)),
                ("search", django.contrib.postgres.search.SearchVectorField(null=True)),
                (
                    "collection",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="data",
                        to="flow.collection",
                    ),
                ),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_data", "Can view data"),
                    ("edit_data", "Can edit data"),
                    ("share_data", "Can share data"),
                    ("owner_data", "Is owner of the data"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="DescriptorSchema",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.fields.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                ("description", models.TextField(blank=True)),
                ("schema", models.JSONField(default=list)),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_descriptorschema", "Can view descriptor schema"),
                    ("edit_descriptorschema", "Can edit descriptor schema"),
                    ("share_descriptorschema", "Can share descriptor schema"),
                    ("owner_descriptorschema", "Is owner of the description schema"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Entity",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.fields.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                ("description", models.TextField(blank=True)),
                ("settings", models.JSONField(default=dict)),
                ("descriptor", models.JSONField(default=dict)),
                ("descriptor_dirty", models.BooleanField(default=False)),
                (
                    "tags",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        default=list,
                        size=None,
                    ),
                ),
                ("search", django.contrib.postgres.search.SearchVectorField(null=True)),
                (
                    "type",
                    models.CharField(
                        blank=True, db_index=True, max_length=100, null=True
                    ),
                ),
                ("duplicated", models.DateTimeField(blank=True, null=True)),
                (
                    "collection",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="flow.collection",
                    ),
                ),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "descriptor_schema",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="flow.descriptorschema",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_entity", "Can view entity"),
                    ("edit_entity", "Can edit entity"),
                    ("share_entity", "Can share entity"),
                    ("owner_entity", "Is owner of the entity"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Process",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.fields.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "type",
                    models.CharField(
                        max_length=100,
                        validators=[
                            django.core.validators.RegexValidator(
                                code="invalid_type",
                                message="Type may be alphanumerics separated by colon",
                                regex="^data:[a-z0-9:]+:$",
                            )
                        ],
                    ),
                ),
                (
                    "category",
                    models.CharField(
                        default="Other:",
                        max_length=200,
                        validators=[
                            django.core.validators.RegexValidator(
                                code="invalid_category",
                                message="Category may be alphanumerics separated by colon",
                                regex="^([a-zA-Z0-9]+[:\\-])*[a-zA-Z0-9]+:$",
                            )
                        ],
                    ),
                ),
                (
                    "persistence",
                    models.CharField(
                        choices=[("RAW", "Raw"), ("CAC", "Cached"), ("TMP", "Temp")],
                        default="RAW",
                        max_length=3,
                    ),
                ),
                ("is_active", models.BooleanField(default=True, verbose_name="active")),
                ("description", models.TextField(default="")),
                ("data_name", models.CharField(blank=True, max_length=200, null=True)),
                ("input_schema", models.JSONField(blank=True, default=list)),
                ("output_schema", models.JSONField(blank=True, default=list)),
                (
                    "entity_type",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                (
                    "entity_descriptor_schema",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                (
                    "entity_input",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                ("entity_always_create", models.BooleanField(default=False)),
                ("run", models.JSONField(default=dict)),
                ("requirements", models.JSONField(default=dict)),
                (
                    "scheduling_class",
                    models.CharField(
                        choices=[("IN", "Interactive"), ("BA", "Batch")],
                        default="BA",
                        max_length=2,
                    ),
                ),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_process", "Can view process"),
                    ("share_process", "Can share process"),
                    ("owner_process", "Is owner of the process"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Relation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.fields.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "category",
                    django.contrib.postgres.fields.citext.CICharField(max_length=100),
                ),
                (
                    "unit",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("s", "Second"),
                            ("min", "Minute"),
                            ("hr", "Hour"),
                            ("d", "Day"),
                            ("wk", "Week"),
                        ],
                        max_length=3,
                        null=True,
                    ),
                ),
                (
                    "collection",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="flow.collection",
                    ),
                ),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_relation", "Can view relation"),
                    ("edit_relation", "Can edit relation"),
                    ("share_relation", "Can share relation"),
                    ("owner_relation", "Is owner of the relation"),
                ),
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="RelationType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, unique=True)),
                ("ordered", models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name="Worker",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("started", models.DateTimeField(auto_now=True)),
                ("finished", models.DateTimeField(null=True)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("PR", "Processing data"),
                            ("NR", "Unresponsive"),
                            ("PP", "Preparing data"),
                            ("CM", "Finished"),
                        ],
                        max_length=2,
                    ),
                ),
                (
                    "data",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="worker",
                        to="flow.data",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Storage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    resolwe.flow.models.fields.ResolweSlugField(
                        max_length=100, populate_from="name", unique_with=("version",)
                    ),
                ),
                ("version", versionfield.fields.VersionField(default="0.0.0")),
                ("name", models.CharField(max_length=100)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                ("json", models.JSONField()),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "data",
                    models.ManyToManyField(related_name="storages", to="flow.Data"),
                ),
            ],
            options={
                "get_latest_by": "version",
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Secret",
            fields=[
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "handle",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("value", fernet_fields.fields.EncryptedTextField()),
                ("metadata", models.JSONField(default=dict)),
                ("expires", models.DateTimeField(db_index=True, null=True)),
                (
                    "contributor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="RelationPartition",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "position",
                    models.PositiveSmallIntegerField(
                        blank=True, db_index=True, null=True
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        blank=True, db_index=True, max_length=30, null=True
                    ),
                ),
                (
                    "entity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="flow.entity"
                    ),
                ),
                (
                    "relation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="flow.relation"
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="relation",
            name="entities",
            field=models.ManyToManyField(
                through="flow.RelationPartition", to="flow.Entity"
            ),
        ),
        migrations.AddField(
            model_name="relation",
            name="type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="flow.relationtype"
            ),
        ),
        migrations.CreateModel(
            name="ProcessMigrationHistory",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("migration", models.CharField(db_index=True, max_length=255)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("metadata", models.JSONField(default=dict)),
                (
                    "process",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="migration_history",
                        to="flow.process",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="DataMigrationHistory",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("migration", models.CharField(db_index=True, max_length=255)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("metadata", models.JSONField(default=dict)),
                (
                    "data",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="migration_history",
                        to="flow.data",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="DataDependency",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "kind",
                    models.CharField(
                        choices=[
                            ("io", "Input/output dependency"),
                            ("subprocess", "Subprocess"),
                            ("duplicate", "Duplicate"),
                        ],
                        max_length=16,
                    ),
                ),
                (
                    "child",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="parents_dependency",
                        to="flow.data",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="children_dependency",
                        to="flow.data",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="data",
            name="descriptor_schema",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="flow.descriptorschema",
            ),
        ),
        migrations.AddField(
            model_name="data",
            name="entity",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="data",
                to="flow.entity",
            ),
        ),
    ]



""".. Ignore pydocstyle D400.

==================
Test template tags
==================

"""


def increase(value):
    """Test template tag that returns an increased value."""
    return value + 1


# A dictionary of filters that will be registered.
filters = {
    "increase": increase,
}



from django.core.exceptions import ValidationError
from django.db import migrations

from django.utils.timezone import now

from resolwe.flow.utils import iterate_fields


def rewire_inputs(data_list):
    """Rewire inputs of provided data objects.

    Input parameter is a list of original and copied data object model
    instances: ``[{'original': original, 'copy': copy}]``. This
    function finds which objects reference other objects (in the list)
    on the input and replaces original objects with the copies (mutates
    copies' inputs).

    """
    if len(data_list) < 2:
        return data_list

    mapped_ids = {bundle["original"].id: bundle["copy"].id for bundle in data_list}

    for bundle in data_list:
        updated = False
        copy = bundle["copy"]

        for field_schema, fields in iterate_fields(
            copy.input, copy.process.input_schema
        ):
            name = field_schema["name"]
            value = fields[name]

            if field_schema["type"].startswith("data:") and value in mapped_ids:
                fields[name] = mapped_ids[value]
                updated = True

            elif field_schema["type"].startswith("list:data:") and any(
                [id_ in mapped_ids for id_ in value]
            ):
                fields[name] = [
                    mapped_ids[id_] if id_ in mapped_ids else id_ for id_ in value
                ]
                updated = True

        if updated:
            copy.save()

    return data_list


def copy_permissions(original, duplicate, model, apps):
    """Copy permissions from original object to it's duplicate."""
    UserObjectPermission = apps.get_model("guardian", "UserObjectPermission")
    GroupObjectPermission = apps.get_model("guardian", "GroupObjectPermission")

    kwargs = {
        "permission__content_type__model": model,
        "object_pk": original.id,
    }
    for perm in UserObjectPermission.objects.filter(**kwargs):
        perm.id = None
        perm.object_pk = duplicate.id
        perm.content_object = duplicate
        perm.save(force_insert=True)

    for perm in GroupObjectPermission.objects.filter(**kwargs):
        perm.id = None
        perm.object_pk = duplicate.id
        perm.content_object = duplicate
        perm.save(force_insert=True)


def duplicate_entity(original, apps):
    """Duplicate entity."""
    Entity = apps.get_model("flow", "Entity")

    duplicate = Entity.objects.get(id=original.id)
    duplicate.pk = None
    duplicate.slug = None
    duplicate.collection = None
    duplicate.name = "Copy of {}".format(original.name)
    duplicate.duplicated = now()

    duplicate.save(force_insert=True)

    copy_permissions(original, duplicate, model="entity", apps=apps)

    # Override fields that are automatically set on create.
    duplicate.created = original.created
    duplicate.save()

    # Duplicate entity's data objects.
    bundle = [
        {"original": obj, "copy": duplicate_data(obj, apps)}
        for obj in original.data.all()
    ]
    bundle = rewire_inputs(bundle)
    duplicated_data = [item["copy"] for item in bundle]

    duplicate.data.add(*duplicated_data)

    return duplicate


def duplicate_data(original, apps):
    """Duplicate (make a copy)."""
    DataDependency = apps.get_model("flow", "DataDependency")
    Data = apps.get_model("flow", "Data")

    if original.status not in ["OK", "ER"]:
        raise ValidationError(
            "Data object must have done or error status to be duplicated"
        )

    duplicate = Data.objects.get(id=original.id)
    duplicate.pk = None
    duplicate.slug = None
    duplicate.entity2 = None
    duplicate.collection = None
    duplicate.name = "Copy of {}".format(original.name)
    duplicate.duplicated = now()

    duplicate.save(force_insert=True)

    copy_permissions(original, duplicate, model="data", apps=apps)

    # Override fields that are automatically set on create.
    duplicate.created = original.created
    duplicate.save()

    if original.location:
        original.location.data.add(duplicate)

    duplicate.storages.set(original.storages.all())

    for migration in original.migration_history.order_by("created"):
        migration.pk = None
        migration.data = duplicate
        migration.save(force_insert=True)

    # Inherit existing child dependencies.
    DataDependency.objects.bulk_create(
        [
            DataDependency(
                child=duplicate, parent=dependency.parent, kind=dependency.kind
            )
            for dependency in DataDependency.objects.filter(child=original)
        ]
    )
    # Inherit existing parent dependencies.
    DataDependency.objects.bulk_create(
        [
            DataDependency(
                child=dependency.child, parent=duplicate, kind=dependency.kind
            )
            for dependency in DataDependency.objects.filter(parent=original)
        ]
    )

    return duplicate


def main(apps, schema_editor):
    """Reassign Data, Entities and Collections relations from m2m to m2o fields."""
    Data = apps.get_model("flow", "Data")
    Entity = apps.get_model("flow", "Entity")

    # 1: Migrate entities from multiple collections.
    for entity in Entity.objects.iterator():

        count = entity.collections.count()
        if count == 1:
            entity.collection = entity.collections.first()
            entity.save()
        elif count > 1:
            collections = entity.collections.all()
            for i, collection in enumerate(collections, start=1):
                # For the last collection, we can keep existing one.
                if i == count:
                    duplicate = entity
                else:
                    duplicate = duplicate_entity(entity, apps)

                duplicate.collection = collection
                duplicate.save()
                for datum in duplicate.data.all():
                    datum.collection = collection
                    datum.save()

    # 2: Migrate data from multiple collections.
    for data in Data.objects.iterator():

        count = data.collection_set.count()
        if count == 1:
            data.collection = data.collection_set.first()
            data.save()
        elif count > 1:
            collections = data.collection_set.all()
            for i, collection in enumerate(collections, start=1):
                # For the last collection, we can keep existing one.
                if i == count:
                    duplicate = data
                else:
                    duplicate = duplicate_data(data, apps)

                duplicate.collection = collection
                duplicate.save()

    # 3: Migrate data from multiple entities.
    for data in Data.objects.iterator():

        count = data.entity_set.count()
        if count == 1:
            data.entity2 = data.entity_set.first()
            data.save()
        elif count > 1:
            entities = data.entity_set.all()
            for i, entity in enumerate(entities, start=1):
                # For the last entity, we can keep existing one.
                if i == count:
                    duplicate = data
                else:
                    duplicate = duplicate_data(data, apps)

                duplicate.entity2 = entity
                duplicate.save()


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0036_add_m2o_fields"),
        ("contenttypes", "0002_remove_content_type_name"),
        ("guardian", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(main, atomic=True),
    ]



"""Storage cleanup."""
import logging
from typing import Optional

from django.db import transaction

from resolwe.storage.models import FileStorage, StorageLocation
from resolwe.utils import BraceMessage as __

logger = logging.getLogger(__name__)


class Cleaner:
    """Remove unreferenced data.

    Remove data from StorageLocation objects that are no longer referenced
    by any Data object.
    """

    def _cleanup(self, storage_location: StorageLocation):
        """Delete data from StorageLocation object."""
        # Make sure this will get writen to the database.
        # Problem is we are inside transaction so what happens if the
        # connection to the database fails or database crashed or...
        connector = storage_location.connector
        if connector is None:
            logger.error(
                __(
                    "Unable to cleanup StorageLocation {}: connector not found.",
                    storage_location.id,
                )
            )
            return
        try:
            # Delete the storage location(files will be also removed).
            logger.info(__("Deleting StorageLocation {}.", storage_location.pk))
            storage_location.delete()

        except Exception:
            logger.exception(
                __(
                    "Exception deleting StorageLocation {}.",
                    storage_location.id,
                )
            )

    def _process_file_storage(self, file_storage: FileStorage):
        """Delete all data from FileStorage object."""
        # Do not remove locked StorageLocation.
        logger.info(__("Processing FileStorage {}.", file_storage.pk))
        for storage_location in file_storage.storage_locations.filter(
            status=StorageLocation.STATUS_DELETING
        ):
            self._cleanup(storage_location)

        if file_storage.storage_locations.count() == 0:
            logger.info(__("Deleting FileStorage {}.", file_storage.pk))
            file_storage.delete()

    def process(self, file_storage_id: Optional[int] = None):
        """Process objects to clean.

        When file_storage is not None process only that object.
        """

        qset = FileStorage.objects.all()
        if file_storage_id is not None:
            logger.info(__("Starting processing for FileStorage {}.", file_storage_id))
            qset = qset.filter(pk=file_storage_id)
        else:
            logger.info("Starting processing")
        for file_storage in qset.filter(data__isnull=True).iterator():
            # Set applicable storage locations to deleting.
            StorageLocation.all_objects.unreferenced_locations().filter(
                file_storage=file_storage
            ).update(status=StorageLocation.STATUS_DELETING)
            with transaction.atomic():
                q_set = FileStorage.objects.filter(
                    id=file_storage.id
                ).select_for_update(skip_locked=True)
                # The FileStorage object is locked or deleted, skip processing.
                if not q_set.exists():
                    logger.debug(
                        "Skipping locked or deleted FileStorage {}.", file_storage.id
                    )
                    continue
                self._process_file_storage(q_set.first())

        logger.info("Finished processing")



"""Kubernetes init container initialization script.

1. Fix permission (make genialis user owner) on subdirectories on the EBS
   volume used for local processing.
2. Transfer missing data to the input EBS volume.
"""

import asyncio
import logging
import os
import shutil
import sys
from collections import defaultdict
from contextlib import suppress
from typing import Any

import zmq
import zmq.asyncio
from executors import constants, global_settings
from executors.connectors import Transfer, connectors
from executors.connectors.baseconnector import BaseStorageConnector
from executors.connectors.exceptions import DataTransferError
from executors.connectors.utils import paralelize
from executors.protocol import ExecutorFiles
from executors.socket_utils import BaseCommunicator, BaseProtocol, Message, PeerIdentity
from executors.zeromq_utils import ZMQCommunicator

from .transfer import transfer_data

logging.basicConfig(
    stream=sys.stdout,
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)
logger = logging.getLogger(__name__)


LISTENER_IP = os.getenv("LISTENER_SERVICE_HOST", "127.0.0.1")
LISTENER_PORT = os.getenv("LISTENER_SERVICE_PORT", "53893")
LISTENER_PROTOCOL = os.getenv("LISTENER_PROTOCOL", "tcp")

DATA_ID = int(os.getenv("DATA_ID", "-1"))


GENIALIS_UID = int(os.environ.get("GENIALIS_UID", 0))
GENIALIS_GID = int(os.environ.get("GENIALIS_GID", 0))
MOUNTED_CONNECTORS = [
    name for name in os.environ["MOUNTED_CONNECTORS"].split(",") if name
]

DOWNLOAD_WAITING_TIMEOUT = 60  # in seconds
RETRIES = 5

# Configure container logger. All logs are output to stdout for further
# processing.
# The log level defaults to debug except for boto and google loggers, which
# is set to warning due to them being extremely verbose in the debug mode.

LOG_LEVEL = int(os.getenv("LOG_LEVEL", logging.DEBUG))
BOTO_LOG_LEVEL = int(os.getenv("BOTO_LOG_LEVEL", logging.WARNING))
GOOGLE_LOG_LEVEL = int(os.getenv("GOOGLE_LOG_LEVEL", logging.WARNING))

logging.basicConfig(
    stream=sys.stdout,
    level=LOG_LEVEL,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)

for boto_logger in ["botocore", "boto3", "s3transfer", "urllib3"]:
    logging.getLogger(boto_logger).setLevel(BOTO_LOG_LEVEL)

for google_logger in ["google"]:
    logging.getLogger(google_logger).setLevel(GOOGLE_LOG_LEVEL)


class PreviousDataExistsError(Exception):
    """Raised if data from previous run exists."""


async def transfer_inputs(communicator: BaseCommunicator, missing_data: dict):
    """Transfer missing input data.

    :raises DataTransferError: on failure.
    """
    inputs_connector = connectors["_input"]
    # Group files by connectors. So we can transfer files from single connector
    # in parallel. We could also transfer files belonging to different
    # connectors in parallel but this could produce huge number of threads,
    # since S3 uses multiple threads to transfer single file.
    objects_to_transfer: dict[str, list[dict[str, str]]] = defaultdict(list)
    for base_url, missing_item in missing_data.items():
        for file in missing_item["files"]:
            file.update({"from_base_url": base_url, "to_base_url": base_url})
        objects_to_transfer[missing_item["from_connector"]] += missing_item["files"]

    try:
        for connector_name in objects_to_transfer:
            min_threads = 5
            max_threads = 20
            # Use from min_threads to max_threads threads. Assume each thread
            # can handle at least 100 files in reasonable time.
            files_count = len(objects_to_transfer[connector_name])
            threads = max(min_threads, min(files_count // 100, max_threads))

            await download_to_location(
                objects_to_transfer[connector_name],
                connectors[connector_name],
                inputs_connector,
                threads,
            )
    except:
        error_message = (
            f"Preparing inputs {objects_to_transfer[connector_name]} from "
            f"connector {connector_name} failed."
        )
        await error(error_message, communicator)
        raise DataTransferError(error_message)


async def download_to_location(
    files: list[dict[str, str]],
    from_connector: BaseStorageConnector,
    to_connector: BaseStorageConnector,
    max_threads: int = 5,
):
    """Download missing paths.

    :raises DataTransferError: on failure.
    """
    logger.info(
        "Transfering %d files from '%s' to '%s' using %d threads.",
        len(files),
        from_connector,
        to_connector,
        max_threads,
    )
    transfer = Transfer(from_connector, to_connector)
    # Start futures and evaluate their results. If exception occured it will
    # be re-raised.
    for future in paralelize(
        objects=files,
        worker=lambda objects: transfer.transfer_chunk(None, objects),
        max_threads=max_threads,
    ):
        future.result()


def check_for_previous_data():
    """Check if previous data exists.

    This method must check processing volume and all mounted shared volumes
    for stdout file.

    :raises PreviousDataExistsError: when data from previous run exists.
    """
    paths_to_check = [
        constants.PROCESSING_VOLUME / global_settings.LOCATION_SUBPATH / "stdout.txt"
    ]
    data_storage = global_settings.SETTINGS["FLOW_STORAGE"]["data"]
    for connector_name in data_storage["connectors"]:
        connector = connectors[connector_name]
        if connector.mountable:
            paths_to_check.append(
                connector.path / global_settings.LOCATION_SUBPATH / "stdout.txt"
            )
    if any(path.is_file() for path in paths_to_check):
        raise PreviousDataExistsError(
            "File 'stdout.txt' exists from previous run, aborting processing."
        )


def prepare_volumes():
    """Create necessary folders and set permissions.

    Prepare a folder LOCATION_SUBPATH inside processing volume and set the
    applicable owner and permissions.
    """
    check_for_previous_data()
    for volume in [constants.PROCESSING_VOLUME, constants.INPUTS_VOLUME]:
        if volume.is_dir():
            logger.debug("Preparing %s.", volume)
            directory = volume / global_settings.LOCATION_SUBPATH
            directory_mode = global_settings.SETTINGS.get("FLOW_EXECUTOR", {}).get(
                "DATA_DIR_MODE", 0o755
            )
            directory.mkdir(mode=directory_mode, exist_ok=True)
            with suppress(PermissionError):
                shutil.chown(directory, GENIALIS_UID, GENIALIS_GID)


def _get_communicator() -> ZMQCommunicator:
    """Connect to the listener."""
    zmq_context = zmq.asyncio.Context.instance()
    zmq_socket = zmq_context.socket(zmq.DEALER)
    zmq_socket.setsockopt(zmq.IDENTITY, str(DATA_ID).encode())
    connect_string = f"{LISTENER_PROTOCOL}://{LISTENER_IP}:{LISTENER_PORT}"
    logger.debug("Opening connection to %s", connect_string)
    zmq_socket.connect(connect_string)
    return ZMQCommunicator(zmq_socket, "init_container <-> listener", logger)


def initialize_secrets(secrets: dict[str, Any]):
    """Initialize secrets."""
    for file_name, content in secrets.items():
        with (constants.SECRETS_VOLUME / file_name).open("wt") as stream:
            stream.write(content)


class InitProtocol(BaseProtocol):
    """Protocol class."""

    async def post_terminate(self, message: Message, identity: PeerIdentity):
        """Handle post-terminate command."""
        logger.debug("Init container received terminate request, terminating.")
        await error("Init container received terminating request.", self.communicator)
        for task in asyncio.all_tasks():
            task.cancel()

    async def transfer_missing_data(self):
        """Transfer missing data.

        :raises DataTransferError: when data transfer error occurs.
        """
        await self.communicator.update_status("PP")
        missing_data = (await self.communicator.missing_data_locations("")).message_data

        await self.communicator.init_suspend_heartbeat("")

        to_filesystem = {
            url: entry | {"url": url}
            for url, entry in missing_data.items()
            if "to_connector" in entry
        }
        to_inputs = {
            url: entry
            for url, entry in missing_data.items()
            if "to_connector" not in entry
        }

        if to_inputs and "_input" not in connectors:
            raise RuntimeError("No inputs volume is defined.")
        if to_filesystem and to_inputs:
            raise RuntimeError(
                "Can only transfer data to input volume or shared storage, not both."
            )

        if to_inputs:
            transfering_coroutine = transfer_inputs(self.communicator, to_inputs)
        else:
            transfering_coroutine = transfer_data(
                self.communicator, list(to_filesystem.values())
            )
        await transfering_coroutine


async def error(error_message: str, communicator: BaseCommunicator):
    """Error occured inside container.

    Send the error and terminate the process.
    """
    with suppress(Exception):
        await communicator.process_log({"error": error_message})


def modify_connector_settings():
    """Modify mountpoints and add processing and input connectors.

    The value of the key 'path' in config dictionary points to the path on the
    worker node. It has to be remaped to a path inside container.
    Also add processing and input connector settings.
    """
    connector_settings = global_settings.SETTINGS["STORAGE_CONNECTORS"]
    storage_settings = global_settings.SETTINGS["FLOW_STORAGE"]
    connector_storage = {
        connector_name: storage_name
        for storage_name in storage_settings
        for connector_name in storage_settings[storage_name]["connectors"]
    }

    # Point connector path to the correct mountpoint.
    for connector_name in MOUNTED_CONNECTORS:
        storage_name = connector_storage[connector_name]
        connector_settings[connector_name]["config"][
            "path"
        ] = f"/{storage_name}_{connector_name}"

    connector_settings["_processing"] = {
        "connector": "executors.connectors.localconnector.LocalFilesystemConnector",
        "config": {"path": constants.PROCESSING_VOLUME},
    }
    if constants.INPUTS_VOLUME_NAME in global_settings.SETTINGS["FLOW_VOLUMES"]:
        connector_settings["_input"] = {
            "connector": "executors.connectors.localconnector.LocalFilesystemConnector",
            "config": {"path": constants.INPUTS_VOLUME},
        }


async def main():
    """Start the main program.

    :raises RuntimeError: when runtime error occurs.
    :raises asyncio.exceptions.CancelledError: when task is terminated.
    """
    communicator = _get_communicator()
    protocol = InitProtocol(communicator, logger)
    communicate_task = asyncio.ensure_future(protocol.communicate())

    # Initialize settings constants by bootstraping.
    response = await communicator.send_command(
        Message.command("bootstrap", (DATA_ID, "init"))
    )
    global_settings.initialize_constants(DATA_ID, response.message_data)
    initialize_secrets(response.message_data[ExecutorFiles.SECRETS_DIR])
    modify_connector_settings()
    connectors.recreate_connectors()

    try:
        prepare_volumes()
        await protocol.transfer_missing_data()
    except PreviousDataExistsError:
        logger.warning("Previous run data exists, aborting processing.")
        raise
    except Exception as exception:
        message = f"Unexpected exception in init container: {exception}."
        logger.exception(message)
        await error(message, communicator)
        raise
    finally:
        protocol.stop_communicate()
        with suppress(asyncio.TimeoutError):
            await asyncio.wait_for(communicate_task, timeout=10)


if __name__ == "__main__":
    logger.debug("Starting the main program.")
    try:
        asyncio.run(main())
    except PreviousDataExistsError:
        sys.exit(2)
    except:
        sys.exit(1)



# Generated by Django 2.2.10 on 2020-05-12 08:44
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("storage", "0002_create_filestorage_objects"),
    ]

    operations = [
        migrations.AddField(
            model_name="referencedpath",
            name="awss3etag",
            field=models.CharField(default="etag", max_length=50),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="referencedpath",
            name="crc32c",
            field=models.CharField(default="crc32c", max_length=8),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="referencedpath",
            name="md5",
            field=models.CharField(default="md5", max_length=32),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.2 on 2019-04-24 07:21

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0033_move_purged"),
    ]

    operations = [
        migrations.AlterField(
            model_name="data",
            name="process_error",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=255), default=list, size=None
            ),
        ),
        migrations.AlterField(
            model_name="data",
            name="process_info",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=255), default=list, size=None
            ),
        ),
        migrations.AlterField(
            model_name="data",
            name="process_warning",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=255), default=list, size=None
            ),
        ),
    ]



""".. Ignore pydocstyle D400.

=================================
Standalone Executor Contact Point
=================================

Command to run on local machine::

    ./manage.py runlistener

"""

import asyncio
from signal import SIGINT, SIGTERM

from django.core.management.base import BaseCommand

from resolwe.flow.managers.listener import ExecutorListener


class Command(BaseCommand):
    """Run the executor listener."""

    help = "Run the standalone manager contact point for executors."

    def add_arguments(self, parser):
        """Add command arguments."""
        super().add_arguments(parser)
        parser.add_argument(
            "--clear-queue",
            action="store_true",
            help="Consume and ignore any outstanding messages in the result queue on startup.",
        )

    def handle(self, *args, **kwargs):
        """Run the executor listener. This method never returns."""

        async def _runner():
            """Run the listener instance."""

            listener = ExecutorListener()

            async def _killer():
                """Kill the listener on receipt of a signal."""
                listener.terminate()

            loop.add_signal_handler(SIGINT, lambda: asyncio.ensure_future(_killer()))
            loop.add_signal_handler(SIGTERM, lambda: asyncio.ensure_future(_killer()))
            async with listener:
                await listener.should_stop.wait()

        loop = asyncio.new_event_loop()
        loop.run_until_complete(_runner())
        loop.close()



from django.db import migrations


def delete_nonexisting_paths(apps, schema_editor):
    """Remove referenced paths that point to exported files.

    Due to a bug in communication container references to all files are stored
    in the datase, not only to files in 'data' storage.

    Specifically the database contains references to files in the 'upload'
    storage, which are named 'export_#hex', where #hex is a 32 characters long
    hexadecimal stored as string.

    This scripts detects such paths and removes them if they do not exist in
    the referenced storage location (to make sure no legit similar named file
    is deleted).
    """
    ReferencedPath = apps.get_model("storage", "ReferencedPath")

    # Remove non-existing exported files paths from storage locations.
    for path in ReferencedPath.objects.filter(path__regex="export_[0-9a-f]{32}$"):
        for storage_location in path.storage_locations.all():
            connector = storage_location.connector
            url = f"{storage_location.file_storage_id}/{path.path}"
            if not connector.exists(url):
                storage_location.files.remove(path)

    # Delete all non-referenced paths.
    ReferencedPath.objects.filter(storage_locations__isnull=True).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("storage", "0008_accesslog_cause"),
    ]

    operations = [
        migrations.RunPython(delete_nonexisting_paths),
    ]



""".. Ignore pydocstyle D400.

==================
Resolwe Test Cases
==================

.. autoclass:: resolwe.test.TestCaseHelpers
    :members:

.. autoclass:: resolwe.test.TransactionTestCase
    :members:

.. autoclass:: resolwe.test.TestCase
    :members:

.. automodule:: resolwe.test.testcases.process

.. automodule:: resolwe.test.testcases.api

"""

import os
import shutil

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.test import SimpleTestCase as DjangoSimpleTestCase
from django.test import TestCase as DjangoTestCase
from django.test import TransactionTestCase as DjangoTransactionTestCase

from resolwe.storage import settings as storage_settings
from resolwe.storage.connectors import connectors


class TestCaseHelpers(DjangoSimpleTestCase):
    """Mixin for test case helpers."""

    def _pre_setup(self, *args, **kwargs):
        # NOTE: This is a work-around for Django issue #10827
        # (https://code.djangoproject.com/ticket/10827) that clears the
        # ContentType cache before permissions are setup.
        ContentType.objects.clear_cache()
        super()._pre_setup(*args, **kwargs)

    def _get_testing_directories(self):
        """Get the testing directories."""
        dirs = [connector.path for connector in connectors.for_storage("data")]
        dirs += [connector.path for connector in connectors.for_storage("upload")]
        dirs += [
            storage_settings.FLOW_VOLUMES[volume_name]["config"]["path"]
            for volume_name in ["processing", "input"]
            if volume_name in storage_settings.FLOW_VOLUMES
        ]
        return dirs

    def _clean_up(self):
        """Clean up after test."""
        if not self._keep_data:
            # Do delete this here. See comment below near the makedirs
            # in setUp.
            for directory in self._get_testing_directories():
                shutil.rmtree(directory, ignore_errors=True)

    def setUp(self):
        """Prepare environment for test."""
        super().setUp()

        # Directories need to be recreated here in case a previous
        # TestCase deleted them. Moving this logic into the test runner
        # and manager infrastructure would not work, because the manager
        # and listener can't know where the testcase boundaries are,
        # they just see a series of data objects; deleting too soon
        # might cause problems for some tests. The runner does not have
        # any code between tests, so could only remove data at the very
        # end, by which time it's already too late, since some tests may
        # deal specifically with the purging functionality and should
        # start in a clean environment, without the sediment
        # (e.g. jsonout.txt, stdout.txt) from previous tests.
        for directory in self._get_testing_directories():
            os.makedirs(directory, exist_ok=True)
        self._keep_data = settings.FLOW_MANAGER_KEEP_DATA

        self.addCleanup(self._clean_up)

    def keep_data(self, mock_purge=True):
        """Do not delete output files after tests."""
        self.fail(
            "*ERROR* TestCaseHelpers.keep_data() is deprecated and does not work anymore.\n"
            "Using it will result in attribute errors in the future.\n"
            "Please use the command line options --keep-data and --no-mock-purge instead.\n"
        )

    def assertAlmostEqualGeneric(self, actual, expected, msg=None):
        """Assert almost equality for common types of objects.

        This is the same as :meth:`~unittest.TestCase.assertEqual`, but using
        :meth:`~unittest.TestCase.assertAlmostEqual` when floats are encountered
        inside common containers (currently this includes :class:`dict`,
        :class:`list` and :class:`tuple` types).

        :param actual: object to compare
        :param expected: object to compare against
        :param msg: optional message printed on failures
        """
        self.assertEqual(type(actual), type(expected), msg=msg)

        if isinstance(actual, dict):
            self.assertCountEqual(actual.keys(), expected.keys(), msg=msg)
            for key in actual.keys():
                self.assertAlmostEqualGeneric(actual[key], expected[key], msg=msg)
        elif isinstance(actual, (list, tuple)):
            for actual_item, expected_item in zip(actual, expected):
                self.assertAlmostEqualGeneric(actual_item, expected_item, msg=msg)
        elif isinstance(actual, float):
            self.assertAlmostEqual(actual, expected, msg=msg)
        else:
            self.assertEqual(actual, expected, msg=msg)


class TransactionTestCase(TestCaseHelpers, DjangoTransactionTestCase):
    """Base class for writing Resolwe tests not enclosed in a transaction.

    It is based on Django's :class:`~django.test.TransactionTestCase`.
    Use it if you need to access the test's database from another
    thread/process.

    """

    def setUp(self):
        """Initialize test data."""
        super().setUp()

        user_model = get_user_model()
        self.admin = user_model.objects.create_superuser(
            username="admin",
            email="admin@test.com",
            password="admin",
            first_name="James",
            last_name="Smith",
        )
        self.contributor = user_model.objects.create_user(
            username="contributor",
            email="contributor@test.com",
            first_name="Joe",
            last_name="Miller",
        )
        self.user = user_model.objects.create_user(
            username="normal_user",
            email="user@test.com",
            first_name="John",
            last_name="Williams",
        )

        self.group = Group.objects.create(name="Users")
        self.group.user_set.add(self.user)


class TestCase(TransactionTestCase, DjangoTestCase):
    """Base class for writing Resolwe tests.

    It is based on :class:`~resolwe.test.TransactionTestCase` and
    Django's :class:`~django.test.TestCase`.
    The latter encloses the test code in a database transaction that is
    rolled back at the end of the test.

    """



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-21 09:56
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0002_set_process_owners"),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.CreateModel(
                    name="DataDependency",
                    fields=[
                        (
                            "id",
                            models.AutoField(
                                auto_created=True,
                                primary_key=True,
                                serialize=False,
                                verbose_name="ID",
                            ),
                        ),
                    ],
                    options={
                        "db_table": "flow_data_parents",
                    },
                ),
                migrations.AlterField(
                    model_name="data",
                    name="parents",
                    field=models.ManyToManyField(
                        related_name="children",
                        through="flow.DataDependency",
                        to="flow.Data",
                    ),
                ),
                migrations.AddField(
                    model_name="datadependency",
                    name="child",
                    field=models.ForeignKey(
                        db_column="from_data_id",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="flow.Data",
                    ),
                ),
                migrations.AddField(
                    model_name="datadependency",
                    name="parent",
                    field=models.ForeignKey(
                        db_column="to_data_id",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="flow.Data",
                    ),
                ),
            ]
        )
    ]



# Generated by Django 2.2.9 on 2020-04-14 21:44
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0043_full_text_search"),
    ]
    operations = [
        migrations.AlterField(
            model_name="data",
            name="location",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="storage.FileStorage",
                related_name="data",
            ),
        ),
        migrations.DeleteModel(
            name="DataLocation",
        ),
        migrations.AlterField(
            model_name="data",
            name="status",
            field=models.CharField(
                choices=[
                    ("UP", "Uploading"),
                    ("RE", "Resolving"),
                    ("WT", "Waiting"),
                    ("PP", "Preparing"),
                    ("PR", "Processing"),
                    ("OK", "Done"),
                    ("ER", "Error"),
                    ("DR", "Dirty"),
                ],
                default="RE",
                max_length=2,
            ),
        ),
    ]



# Generated by Django 3.1.7 on 2021-11-04 11:13

from django.db import migrations
import django.db.models.manager


class Migration(migrations.Migration):

    dependencies = [
        ("permissions", "0002_migrate_old_permissions"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="permissionmodel",
            managers=[
                ("all_objects", django.db.models.manager.Manager()),
            ],
        ),
    ]



""".. Ignore pydocstyle D400.

=============================
Validate permission hierarchy
=============================

When data/entity object O is contained in container (entity/collection) C it
must inherit its permissions. This is implemented by connecting all objects in
a collection to the same PermissionGroup. However, there is no guarantee that
objects in the collection belong to the same permission group.

This command checks that for every container all objects it point to the same
PermissionGroup object.
"""
from django.core.management.base import BaseCommand

from resolwe.flow.models import Collection, Entity


class Command(BaseCommand):
    """Validate permissions hierarchy."""

    help = "Validate permissions hierarchy."

    def validate_collection(self, collection: Collection) -> bool:
        """Validate collection.

        Check that all objects inside the collection point to the same
        PermissionGroup object.
        """
        return any(
            queryset.exclude(permission_group=collection.permission_group)
            for queryset in (collection.data, collection.entity_set)
        )

    def validate_entity(self, entity: Entity) -> bool:
        """Validate entity.

        Check that all objects inside the collection point to the same
        PermissionGroup object.
        """
        return bool(entity.date.exclude(permission_group=entity.permission_group))

    def add_arguments(self, parser):
        """Command arguments."""
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "-c",
            "--collection-slug",
            dest="collection_slug",
            help="Check only collection with the given slug.",
        )
        group.add_argument(
            "-e",
            "--entity-slug",
            dest="entity_slug",
            help="Check only entity with the given slug.",
        )

    def handle(self, **options):
        """Command handle."""
        collection_queryset = Collection.objects.none()
        entity_queryset = Entity.objects.none()
        if options["collection_slug"]:
            collection_queryset = collection_queryset.filter(
                slug=options["collection_slug"]
            )
        elif options["entity_slug"]:
            entity_queryset = entity_queryset.filter(slug=options["entity_slug"])
        else:
            Collection.objects.all()
            entity_queryset = Entity.objects.filter(collection__isnull=False)

        invalid_collections = [
            collection
            for collection in collection_queryset
            if not self.validate_collection(collection)
        ]
        invalid_entities = [
            entity for entity in entity_queryset if not self.validate_entity(entity)
        ]
        for collection in invalid_collections:
            self.stderr.write(
                f"Permission hierarchy mismatch in collection {collection.slug}"
            )

        for entity in invalid_entities:
            self.stderr.write(f"Permission hierarchy mismatch in entity {entity.slug}")

        if not (invalid_collections or invalid_entities):
            self.stdout.write("Check completed successfully.")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-02-26 04:06
from __future__ import unicode_literals
from collections import defaultdict
from typing import List, Tuple

from django.contrib.postgres.aggregates import ArrayAgg
from django.db import migrations, models

from resolwe.permissions.models import Permission


def _get_permissions(queryset: models.QuerySet, apps) -> Tuple[List, List]:
    """Get all the user and group permissions for given ctype.

    :returns: dictionary with keys 'user' and 'group'. Each contains a
    dictionary with object primary ids as keys and list of tuples of the form
    ({user/group_id}, {permission}) as values.
    """
    GroupObjectPermission = apps.get_model("guardian", "GroupObjectPermission")
    UserObjectPermission = apps.get_model("guardian", "UserObjectPermission")
    ContentType = apps.get_model("contenttypes", "ContentType")

    ctype = ContentType.objects.get_for_model(queryset.model)

    def permission_from_strings(permission_names: List[str]) -> Permission:
        """Return the top permission from permission string.

        The permission strings are in the form '{permission_name}_{model}', for
        instance 'view_data'.
        """
        return max(
            Permission.from_name(permission_name.split("_")[0])
            for permission_name in permission_names
        )

    permission_models = {"user": UserObjectPermission, "group": GroupObjectPermission}

    perms = dict()
    for permission_for in permission_models:
        perms[permission_for] = (
            permission_models[permission_for]
            .objects.filter(content_type=ctype)
            .values("object_pk", entity_id=models.F(f"{permission_for}_id"))
            .annotate(permission=ArrayAgg("permission__codename"))
        )

    permission_maps = {"user": defaultdict(list), "group": defaultdict(list)}
    for permission_for in permission_models:
        for perm in perms[permission_for]:
            permission_maps[permission_for][int(perm["object_pk"])].append(
                (int(perm["entity_id"]), permission_from_strings(perm["permission"]))
            )
    return permission_maps


def migrate_old_permissions(apps, schema_editor):
    """Migrate old Guardian permissions to new permission architecture."""

    def process_queryset(queryset: models.QuerySet):
        """Create PermissionModel instances for the queryset."""

        permissions = _get_permissions(queryset, apps)
        for obj in queryset.iterator():
            if obj.permission_group is None:
                obj.permission_group = PermissionGroup.objects.create()
                obj.save()
            for entity_type in ["user", "group"]:
                for entity_id, permission in permissions[entity_type][obj.id]:
                    permission_models.append(
                        PermissionModel(
                            **{
                                "value": permission.value,
                                "permission_group": obj.permission_group,
                                f"{entity_type}_id": entity_id,
                            }
                        )
                    )

    Data = apps.get_model("flow", "Data")
    Entity = apps.get_model("flow", "Entity")
    Collection = apps.get_model("flow", "Collection")
    DescriptorSchema = apps.get_model("flow", "DescriptorSchema")
    Process = apps.get_model("flow", "Process")

    PermissionGroup = apps.get_model("permissions", "PermissionGroup")
    PermissionModel = apps.get_model("permissions", "PermissionModel")

    # All PermissionModel instances that will be bulk created at the end.
    permission_models = []

    # First process Collections and all its contents.
    process_queryset(Collection.objects.all())
    for collection in Collection.objects.all():
        collection.entity_set.update(permission_group_id=collection.permission_group_id)
        collection.data.update(permission_group_id=collection.permission_group_id)
        collection.relation_set.update(
            permission_group_id=collection.permission_group_id
        )

    # Now process entities without collection, that is entities with permission
    # group set to null in the database.
    remaining_entities = Entity.objects.filter(permission_group__isnull=True)
    # A list is needed of filter bellow will always return empty set
    remaining_entities_list = list(remaining_entities)
    process_queryset(remaining_entities)
    for entity in remaining_entities_list:
        entity.data.update(permission_group_id=entity.permission_group_id)

    # Now process remaining data objects, that is data objects with permission
    # group set to null in the database.
    remaining_data = Data.objects.filter(permission_group__isnull=True)
    process_queryset(remaining_data)

    # Now process descriptor schemas.
    process_queryset(DescriptorSchema.objects.all())

    # Now process processes.
    process_queryset(Process.objects.all())

    # Now bulk create all permission models.
    PermissionModel.objects.bulk_create(permission_models)


class Migration(migrations.Migration):

    dependencies = [
        ("permissions", "0001_initial"),
        ("flow", "0003_add_permission_group"),
    ]

    operations = [
        migrations.RunPython(
            migrate_old_permissions, reverse_code=migrations.RunPython.noop
        ),
    ]



# Generated by Django 2.2.9 on 2020-01-10 16:44
import os

import django.contrib.postgres.indexes
import django.contrib.postgres.search
from django.contrib.postgres.operations import TrigramExtension
from django.db import connection, migrations, models


def load_triggers(apps, schema_editor):
    file_names = [
        "utils.sql",
        "triggers_collection.sql",
        "triggers_entity.sql",
        "triggers_data.sql",
    ]
    with connection.cursor() as c:
        for file_name in file_names:
            file_path = os.path.join(os.path.dirname(__file__), file_name)
            with open(file_path) as fh:
                sql_statement = fh.read()
            c.execute(sql_statement)


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0042_delete_obsolete_perms"),
    ]

    operations = [
        TrigramExtension(),
        migrations.AddField(
            model_name="collection",
            name="search",
            field=django.contrib.postgres.search.SearchVectorField(null=True),
        ),
        migrations.AddField(
            model_name="data",
            name="search",
            field=django.contrib.postgres.search.SearchVectorField(null=True),
        ),
        migrations.AddField(
            model_name="entity",
            name="search",
            field=django.contrib.postgres.search.SearchVectorField(null=True),
        ),
        migrations.RunPython(load_triggers),
        # Update existing entries.
        migrations.RunSQL("UPDATE flow_collection SET id=id;"),
        migrations.RunSQL("UPDATE flow_entity SET id=id;"),
        migrations.RunSQL("UPDATE flow_data SET id=id;"),
        migrations.AddIndex(
            model_name="collection",
            index=models.Index(fields=["name"], name="idx_collection_name"),
        ),
        migrations.AddIndex(
            model_name="collection",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["name"],
                name="idx_collection_name_trgm",
                opclasses=["gin_trgm_ops"],
            ),
        ),
        migrations.AddIndex(
            model_name="collection",
            index=models.Index(fields=["slug"], name="idx_collection_slug"),
        ),
        migrations.AddIndex(
            model_name="collection",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["tags"], name="idx_collection_tags"
            ),
        ),
        migrations.AddIndex(
            model_name="collection",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["search"], name="idx_collection_search"
            ),
        ),
        migrations.AddIndex(
            model_name="data",
            index=models.Index(fields=["name"], name="idx_data_name"),
        ),
        migrations.AddIndex(
            model_name="data",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["name"], name="idx_data_name_trgm", opclasses=["gin_trgm_ops"]
            ),
        ),
        migrations.AddIndex(
            model_name="data",
            index=models.Index(fields=["slug"], name="idx_data_slug"),
        ),
        migrations.AddIndex(
            model_name="data",
            index=models.Index(fields=["status"], name="idx_data_status"),
        ),
        migrations.AddIndex(
            model_name="data",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["tags"], name="idx_data_tags"
            ),
        ),
        migrations.AddIndex(
            model_name="data",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["search"], name="idx_data_search"
            ),
        ),
        migrations.AddIndex(
            model_name="entity",
            index=models.Index(fields=["name"], name="idx_entity_name"),
        ),
        migrations.AddIndex(
            model_name="entity",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["name"], name="idx_entity_name_trgm", opclasses=["gin_trgm_ops"]
            ),
        ),
        migrations.AddIndex(
            model_name="entity",
            index=models.Index(fields=["slug"], name="idx_entity_slug"),
        ),
        migrations.AddIndex(
            model_name="entity",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["tags"], name="idx_entity_tags"
            ),
        ),
        migrations.AddIndex(
            model_name="entity",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["search"], name="idx_entity_search"
            ),
        ),
        migrations.AddIndex(
            model_name="process",
            index=models.Index(fields=["name"], name="idx_process_name"),
        ),
        migrations.AddIndex(
            model_name="process",
            index=models.Index(fields=["slug"], name="idx_process_slug"),
        ),
        migrations.AddIndex(
            model_name="process",
            index=models.Index(fields=["type"], name="idx_process_type"),
        ),
        migrations.AddIndex(
            model_name="process",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["type"],
                name="idx_process_type_trgm",
                opclasses=["gin_trgm_ops"],
            ),
        ),
    ]



# Generated by Django 3.1.5 on 2021-03-02 08:03

# Migrate from django.contrib.postgres.fields.JSONField to
# django.db.models.JSONField

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0048_worker"),
    ]

    operations = [
        migrations.AlterField(
            model_name="collection",
            name="descriptor",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="collection",
            name="settings",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="data",
            name="descriptor",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="data",
            name="input",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="data",
            name="output",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="datamigrationhistory",
            name="metadata",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="descriptorschema",
            name="schema",
            field=models.JSONField(default=list),
        ),
        migrations.AlterField(
            model_name="entity",
            name="descriptor",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="entity",
            name="settings",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="process",
            name="input_schema",
            field=models.JSONField(blank=True, default=list),
        ),
        migrations.AlterField(
            model_name="process",
            name="output_schema",
            field=models.JSONField(blank=True, default=list),
        ),
        migrations.AlterField(
            model_name="process",
            name="requirements",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="process",
            name="run",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="processmigrationhistory",
            name="metadata",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="secret",
            name="metadata",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="storage",
            name="json",
            field=models.JSONField(),
        ),
    ]



"""Custom database wrapper for PostgreSQL.

Django uses ``LIKE UPPER(%)`` operator for ``icontains``,
``istartswith`` and ``iendswith`` filters. This operator doesn't work
with ``gin_trgm_ops`` index like ``LIKE %s`` operator does - or at least
it requires a separate index built with ``UPPER``.

Better (and faster) solution is to use ``ILIKE`` operator which uses
the same index as ``LIKE``.

Some background for Django's design is available at:
https://code.djangoproject.com/ticket/3575
"""
from django.db.backends.postgresql import base, operations


class DatabaseOperations(
    operations.DatabaseOperations
):  # pylint: disable=abstract-method
    """Database operations class."""

    def lookup_cast(self, lookup_type, internal_type=None):
        """Copy of the original method, but don't use ``UPPER`` for i-lookups."""
        lookup = "%s"

        # Cast text lookups to text to allow things like filter(x__contains=4)
        if lookup_type in (
            "iexact",
            "contains",
            "icontains",
            "startswith",
            "istartswith",
            "endswith",
            "iendswith",
            "regex",
            "iregex",
        ):
            if internal_type in ("IPAddressField", "GenericIPAddressField"):
                lookup = "HOST(%s)"
            else:
                lookup = "%s::text"

        return lookup


class DatabaseWrapper(base.DatabaseWrapper):
    """Database wrapper for PostgreSQL database.

    Inherit from Django's defaulf database wrapper and apply some
    changes to how the SQL queries are composited.
    """

    ops_class = DatabaseOperations

    def __init__(self, *args, **kwargs):
        """Initialize the wrapper."""
        self.operators.update(
            {
                "iexact": "ILIKE %s",
                "icontains": "ILIKE %s",
                "istartswith": "ILIKE %s",
                "iendswith": "ILIKE %s",
            }
        )
        self.pattern_ops.update(
            {
                "iexact": "ILIKE {}",
                "icontains": "ILIKE '%%' || {} || '%%'",
                "istartswith": "ILIKE {} || '%%'",
                "iendswith": "ILIKE '%%' || {}",
            }
        )
        super().__init__(*args, **kwargs)



"""Test resolwe.storage.views."""
import json
from unittest.mock import MagicMock, patch

from django.test import RequestFactory, TestCase

from rest_framework import status

from resolwe.storage.views import UriResolverView


class UriResolverViewTest(TestCase):
    """Test UriResolverView."""

    def setUp(self):
        self.factory = RequestFactory()

    @patch("resolwe.storage.views.UriResolverView._get_datum")
    @patch("resolwe.storage.views.DataBrowseView._get_response")
    def test_post(self, get_response_mock, get_datum_mock):
        """Test post method."""
        get_datum_mock.return_value = MagicMock()
        get_response_mock.side_effect = [
            ("signed_url1", True),
            ("dir_structure", False),
            ("signed_url2", True),
        ]

        uris = [
            "123/file1.txt",
            "456/dir",
            "789/dir/file2.txt",
        ]
        request = self.factory.post("", {"uris": uris}, content_type="application/json")

        response = UriResolverView().post(request)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            json.loads(response.content.decode("utf-8")),
            {
                "123/file1.txt": "signed_url1",
                "456/dir": "dir_structure",
                "789/dir/file2.txt": "signed_url2",
            },
        )

    def test_get(self):
        """Test get method."""
        request = self.factory.get("")
        response = UriResolverView().get(request)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.content.decode("utf-8"), "")



"""Entity viewset."""
from django.db.models import Prefetch

from rest_framework import exceptions
from rest_framework.decorators import action
from rest_framework.response import Response

from resolwe.flow.filters import EntityFilter
from resolwe.flow.models import Collection, DescriptorSchema, Entity
from resolwe.flow.serializers import EntitySerializer

from .collection import BaseCollectionViewSet
from .utils import get_collection_for_user


class EntityViewSet(BaseCollectionViewSet):
    """API view for entities."""

    serializer_class = EntitySerializer
    filter_class = EntityFilter
    qs_collection_ds = DescriptorSchema.objects.select_related("contributor")
    qs_collection = Collection.objects.select_related("contributor")
    qs_collection = qs_collection.prefetch_related(
        "data",
        "entity_set",
        Prefetch("descriptor_schema", queryset=qs_collection_ds),
    )
    qs_descriptor_schema = DescriptorSchema.objects.select_related("contributor")
    queryset = Entity.objects.select_related("contributor").prefetch_related(
        "data",
        Prefetch("collection", queryset=qs_collection),
        Prefetch("descriptor_schema", queryset=qs_descriptor_schema),
    )

    def _get_entities(self, user, ids):
        """Return entities queryset based on provided entity ids."""
        queryset = Entity.objects.filter(id__in=ids).filter_for_user(user)
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Entities with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        return queryset

    @action(detail=False, methods=["post"])
    def move_to_collection(self, request, *args, **kwargs):
        """Move samples from source to destination collection."""
        ids = self.get_ids(request.data)
        dst_collection_id = self.get_id(request.data, "destination_collection")
        dst_collection = get_collection_for_user(dst_collection_id, request.user)

        entity_qs = self._get_entities(request.user, ids)
        entity_qs.move_to_collection(dst_collection)

        return Response()

    # NOTE: This can be deleted when DRF will support select_for_update
    #       on updates and ResolweUpdateModelMixin will use it.
    #       https://github.com/encode/django-rest-framework/issues/4675
    def update(self, request, *args, **kwargs):
        """Update an entity.

        Original queryset produces a temporary database table whose rows
        cannot be selected for an update. As a workaround, we patch
        get_queryset function to return only Entity objects without
        additional data that is not needed for the update.
        """
        orig_get_queryset = self.get_queryset

        def patched_get_queryset():
            """Patched get_queryset method."""
            entity_ids = orig_get_queryset().values_list("id", flat=True)
            return Entity.objects.filter(id__in=entity_ids)

        self.get_queryset = patched_get_queryset

        resp = super().update(request, *args, **kwargs)

        self.get_queryset = orig_get_queryset
        return resp

    @action(detail=False, methods=["post"])
    def duplicate(self, request, *args, **kwargs):
        """Duplicate (make copy of) ``Entity`` models."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        inherit_collection = request.data.get("inherit_collection", False)
        ids = self.get_ids(request.data)
        queryset = Entity.objects.filter(id__in=ids).filter_for_user(request.user)
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Entities with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        duplicated = queryset.duplicate(
            contributor=request.user, inherit_collection=inherit_collection
        )

        serializer = self.get_serializer(duplicated, many=True)
        return Response(serializer.data)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-11-07 08:44
from __future__ import unicode_literals

from django.db import migrations


def populate_entity_type(apps, schema_editor):
    """Populate entity type from attached descriptor schema."""
    Entity = apps.get_model("flow", "Entity")

    for entity in Entity.objects.all():
        if entity.descriptor_schema is not None:
            entity.type = entity.descriptor_schema.slug
            entity.save()


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0024_process_entity_3"),
    ]

    operations = [migrations.RunPython(populate_entity_type)]



"""Urls for Django Rest Framework."""
from django.urls import include, path

from rest_framework import routers

from .flow.views import (
    CollectionViewSet,
    DataViewSet,
    DescriptorSchemaViewSet,
    EntityViewSet,
    ProcessViewSet,
    RelationViewSet,
    StorageViewSet,
)

api_router = routers.DefaultRouter(trailing_slash=False)
api_router.register(r"collection", CollectionViewSet)
api_router.register(r"process", ProcessViewSet)
api_router.register(r"data", DataViewSet)
api_router.register(r"entity", EntityViewSet)
api_router.register(r"relation", RelationViewSet)
api_router.register(r"descriptorschema", DescriptorSchemaViewSet)
api_router.register(r"storage", StorageViewSet)


urlpatterns = [
    path("api/", include((api_router.urls, "resolwe-api"), namespace="resolwe-api")),
]
